import numpy as np

def g(n):
    return 2*n**2

def numberOfOccupationNumbersInShell(n, Z):
    return min(g(n), Z)+1

def stepLength(n, Z):
     stepLength = 1
     for i in range(1,n):
        stepLength *= numberOfOccupationNumbersInShell(i, Z)

     return stepLength

def getLevelVector(x, Z, numLevels):
    levelVector = np.empty(numLevels)

    for n in range(numLevels, 0, -1):
        # calculate current stepLength
        currentStepLength = stepLength(n, Z)

        # get occupation number
        levelVector[n-1] = x // currentStepLength

        # remove contribution of current level
        x -= currentStepLength * levelVector[n - 1]

    return levelVector

def getConfigNumber(levelVector, Z):
    configNumber = 0
    for i in range(0,len(levelVector)):
        configNumber += stepLength(i+1, Z) * levelVector[i]
    return configNumber

# @TODO: write test, BrianMarre, 2021
#testLevelVector = np.array([2,1,1,0,0,0,0,1,0,0]) 352973
#testLevelvector = np.array([2,3,0,0,0,0,0,0,0,0]) 11
#print(getConfigNumber(testLevelVector, 6)) 

filenameOccuppationData= "C:/Users/Kvasir/Desktop/VerificationRunflyCHK/flyspect_data_atomicState.txt"
filenameAtomicStateData="C:/Users/Kvasir/Desktop/VerificationRunflyCHK/data/atomic.inp."

Z=18
numLevels = 10

# load output data:[(<ionizationState>, <Population of lower state>, <Population upper state>, <local state index lower>, <local state index upper>)]
dataOccupation = np.loadtxt(filenameOccuppationData, dtype=[("Iso", 'u1'), ("Pop_lower", 'f8'), ("Pop_upper", 'f4'),
    ("localIndex_lower", 'u4'), ("localIndex_upper", 'u4')], skiprows=26, usecols=[0,8,9,10,11])

# uniqueEntries is dictionary(iso:dictionary(localIndex:Pop)), containing all occupied states once
uniqueEntries = {}
for entry in dataOccupation:
    iso = entry["Iso"]
    uniqueEntriesIonizationState = uniqueEntries.get(iso)

    if uniqueEntriesIonizationState:
        lowerLocalIndex = entry["localIndex_lower"]
        upperLocalIndex = entry["localIndex_upper"]

        uniqueEntryState = uniqueEntriesIonizationState.get(lowerLocalIndex)
        if not uniqueEntryState:
            uniqueEntriesIonizationState[lowerLocalIndex]=entry["Pop_lower"]

        uniqueEntryState = uniqueEntriesIonizationState.get(upperLocalIndex)
        if not uniqueEntryState:
            uniqueEntriesIonizationState[upperLocalIndex]=entry["Pop_upper"]
    else:
        uniqueEntries[iso] = { entry["localIndex_lower"]:entry["Pop_lower"], entry["localIndex_upper"]:entry["Pop_upper"] }

del dataOccupation

#{
# load atomic state data, taken from flylite level reader@Axel
# regex that matches floats and scientific floats
rg_flt = "([-\+[0-9]+\.[0-9]*[Ee]*[\+-]*[0-9]*)"
# read data from file
dtype = np.dtype([
         #      iz                j (in charge_state)      ilev
         ('charge_state', 'int'), ('state_idx', 'int'), ('total_idx', 'int'),
         ('name', '|S8'), # kname(ilev): unique naming of the state
         ('energy', np.float), # elev(ilev): energy from ground state [eV]
         # glev(ilev): number of permutations within state
         #             (spin up/down, angular momentum) [large int]
         ('statistical_weight', np.float),
         ('charge_screened', 'float'), # qlev(ilev)
           ('ionization_potential', 'float'), # evtoip(ilev) [eV]
           ('destruction_rate', 'float'), # gamx: destruction rate [1/s]
         # ncont(i,j): index of its own ionization level
         ('ioniz_lvl_idx', 'int'),
         # mobtot(ilev): maximum principle quantum number with population != 0
         #               (K=1, L=2, ...)
         ('nmax', 'int'),
         # noctot(ilev,k=1-10): shell population
         #   careful, those are not space separated but only aligned to 2 fields
         ('n_1', 'int'),
           ('n_2', 'int'), ('n_3', 'int'), ('n_4', 'int'),
           ('n_5', 'int'), ('n_6', 'int'), ('n_7', 'int'),
           ('n_8', 'int'), ('n_9', 'int'), ('n_10', 'int')
    ])
#}

#load dataStates
dataStates = np.fromregex(filenameAtomicStateData + str(Z),
    "\s*(\d+)\s+(\d+)\s+(\d+)\s+"
    "(........)\s+" +
    rg_flt + "\s+" +
    rg_flt + "\s+" +
    rg_flt + "\s+" + rg_flt + "\s+" + rg_flt + "\s+"
    "(\d+)\s+(\d+)\s+"
    "(\d+)(.\d)(.\d)(.\d)(.\d)(.\d)(.\d)(.\d)(.\d)(.\d)\s*\n",
     dtype=dtype
)

# dictionary(<configNumber>:<number density>)
statePopulations = {}

for state in dataStates:
    uniqueEntriesIonizationState = uniqueEntries.get( state["charge_state"] )

    if uniqueEntriesIonizationState:
        populationState = uniqueEntriesIonizationState.get( state["state_idx"] )

        if populationState:
            levelVector = np.empty(numLevels)

            for i in range(0, numLevels):
                levelVector[i] = state["n_" + str(i+1)]

            configNumber = getConfigNumber(levelVector, Z)
            statePopulations[configNumber]=populationState

print(statePopulations)